from __future__ import print_function

from trakt_tools.core.input import boolean_input
from trakt_tools.models import Backup, Profile
from trakt_tools.tasks.base import Task
from .handlers import CollectionHandler, HistoryHandler, PlaybackHandler, RatingsHandler, WatchlistHandler

from trakt import Trakt
from zipfile import ZipFile, ZIP_DEFLATED
import logging
import os
import shutil

log = logging.getLogger(__name__)


class CreateBackupTask(Task):
    handlers = [
        CollectionHandler,
        HistoryHandler,
        PlaybackHandler,
        RatingsHandler,
        WatchlistHandler
    ]

    def __init__(self, backup_dir, per_page=1000, debug=False, rate_limit=20, assume_yes=False):
        super(CreateBackupTask, self).__init__(
            debug=debug,
            rate_limit=rate_limit
        )

        self.assume_yes = assume_yes
        self.backup_dir = backup_dir
        self.per_page = per_page

    def run(self, token):
        log.debug('run()')

        # Process backup download
        with Trakt.configuration.oauth(token=token):
            return self.process()

    def process(self, profile=None):
        log.debug('process()')

        if not profile:
            print('Requesting profile...')
            profile = Profile.fetch(
                self.per_page,
                self.rate_limit
            )

        if not profile:
            raise Exception('Unable to fetch profile')

        print('Logged in as %r' % profile.username)
        print()

        if not self.assume_yes and not boolean_input('Would you like to continue?', default=True):
            exit(0)

        print()

        # Create backup
        return self.create_backup(profile)

    def create_backup(self, profile):
        backup = Backup.create(self.backup_dir, profile.username)

        # Run handlers
        for handler in self.handlers:
            h = handler()

            if not h.run(backup, profile):
                print('Unable to backup profile, handler %r failed' % h)
                return False

            print()

        # Compress backup
        print('Compressing backup...')

        dest_path = os.path.join(
            self.backup_dir,
            profile.username,
            '%s.zip' % backup.name
        )

        with ZipFile(dest_path, 'w', ZIP_DEFLATED) as archive:
            # Add backup files
            for root, dirs, files in os.walk(backup.path):
                for filename in files:
                    path = os.path.join(root, filename)

                    # Add file to archive
                    archive.write(path, os.path.relpath(path, backup.path))

        # Delete backup directory
        print('Cleaning up...')
        shutil.rmtree(backup.path)

        print('Backup has been saved to: "%s"' % dest_path)
        return True



from .collection import CollectionHandler
from .history import HistoryHandler
from .playback import PlaybackHandler
from .ratings import RatingsHandler
from .watchlist import WatchlistHandler


__all__ = [
    'CollectionHandler',
    'HistoryHandler',
    'PlaybackHandler',
    'RatingsHandler',
    'WatchlistHandler'
]



from __future__ import print_function

from trakt_tools.core.input import boolean_input

from requests import RequestException
from trakt import Trakt
import logging
import pytz
import time

log = logging.getLogger(__name__)


class Profile(object):
    def __init__(self, settings, per_page=None, rate_limit=None):
        self.settings = settings

        self.per_page = per_page
        self.rate_limit = rate_limit

        self._cache = {}
        self._last_request_at = None

    @property
    def username(self):
        if not self.settings:
            return None

        return self.settings.get('user', {}).get('username')

    @property
    def timezone(self):
        name = self.timezone_name

        if not name:
            return None

        try:
            return pytz.timezone(name)
        except Exception:
            log.warn('Unknown account timezone %r, timestamps will be displayed in the GMT/UTC timezone instead', name)
            return None

    @property
    def timezone_name(self):
        if not self.settings:
            return None

        return self.settings.get('account', {}).get('timezone')

    # region Public methods

    # region HTTP

    def clear_cache(self):
        self._cache = None

    def request(self, method, path=None, query=None, data=None):
        # Ignore caching on non-GET requests
        if method != 'GET':
            self._rate_limit()

            # Fire request
            return Trakt.http.request(
                method,
                path=path,
                query=query,
                data=data
            )

        if data:
            raise NotImplementedError("\"data\" parameter not supported")

        # Build cache key
        cache_key = self._build_cache_key(path, query)

        # Check cache for response
        if cache_key not in self._cache:
            self._rate_limit()

            # Send request
            response = Trakt.http.request(
                method,
                path=path,
                query=query
            )

            # Store response is cache (if successful)
            if response.status_code == 200:
                self._cache[cache_key] = response

            return response

        # Return response from cache
        return self._cache[cache_key]

    def delete(self, path=None, query=None, data=None):
        return self.request(
            'DELETE',
            path=path,
            query=query,
            data=data
        )

    def get(self, path=None, query=None):
        return self.request(
            'GET',
            path=path,
            query=query
        )

    def post(self, path=None, query=None, data=None):
        return self.request(
            'POST',
            path=path,
            query=query,
            data=data
        )

    def put(self, path=None, query=None, data=None):
        return self.request(
            'PUT',
            path=path,
            query=query,
            data=data
        )

    # endregion

    def get_pages(self, path=None, query=None):
        page = 1
        page_count = None
        item_count = None

        received_count = 0
        retry_count = 0

        while (page == 1 and page_count is None) or (page_count is not None and page <= page_count):
            page_query = {
                'page': page,
                'limit': self.per_page
            }

            if query:
                page_query.update(query)

            # Request page
            try:
                response = self.get(
                    path=path,
                    query=page_query
                )
            except RequestException as ex:
                self._retry_request(
                    retry_count,
                    prompt=retry_count >= 3,
                    message='Unable to fetch page #%d' % page,
                    reason='exception: %s' % ex.message
                )
                retry_count += 1
                continue

            if response.status_code != 200:
                self._retry_request(
                    retry_count,
                    prompt=retry_count >= 3,
                    message='Unable to fetch page #%d' % page,
                    reason='status: %s' % response.status_code
                )
                retry_count += 1
                continue

            # Valid response returned
            items = response.json()

            # Reset state
            retry_count = 0

            # Increment received count
            received_count += len(items)

            # Update page count
            page_count = int(response.headers['X-Pagination-Page-Count'])

            # Update item count
            item_count = int(response.headers['X-Pagination-Item-Count'])

            # Yield page details
            yield page, page_count, items

            # Increment current `page`
            page += 1

        if received_count != item_count:
            raise Exception("Entire history wasn't retrieved (expected %d item(s), received %d item(s)" % (
                item_count,
                received_count
            ))

    def __repr__(self):
        return '<Profile %r>' % self.username

    # endregion

    # region Private methods

    @staticmethod
    def _build_cache_key(path, query):
        if query:
            path += '?' + ('&'.join([
                key + '=' + str(value)
                for key, value in query.items()
            ]))

        return path

    def _rate_limit(self):
        # Rate-limit requests
        if self.rate_limit is not None and self._last_request_at is not None:
            delay = (60 / self.rate_limit) - (time.time() - self._last_request_at)

            if delay > 0:
                time.sleep(delay)

        self._last_request_at = time.time()

    @staticmethod
    def _retry_request(retry_count, prompt=True, message='Request failed', reason=None):
        # Ask if a retry should be attempted
        if prompt:
            print('%s (%s)' % (
                message,
                reason
            ))

            if boolean_input('Request has failed %d times, retry request?' % retry_count, default=True):
                # Retry request again
                return

            # User cancelled the request
            if reason:
                raise Exception('%s (attempted %d times), %s' % (
                    message,
                    retry_count,
                    reason
                ))

            raise Exception('%s (attempted %d times)' % (
                message,
                retry_count
            ))

        # Retry request
        print('%s, retrying in 5 seconds... (%s)' % (
            message,
            reason
        ))
        time.sleep(5)

    # endregion

    # region Class methods

    @classmethod
    def fetch(cls, per_page, rate_limit):
        settings = Trakt['users/settings'].get()

        if not settings:
            return None

        return cls(
            settings,
            per_page=per_page,
            rate_limit=rate_limit
        )

    # endregion



from __future__ import print_function

import logging

log = logging.getLogger(__name__)


class RatingsHandler(object):
    def run(self, backup, profile):
        print('Ratings')

        # Request ratings
        response = profile.get('/sync/ratings')

        if response.status_code != 200:
            print('Invalid response returned')
            return False

        # Retrieve items
        items = response.json()

        print(' - Received %d item(s)' % len(items))

        # Write ratings to disk
        print(' - Writing to "ratings.json"...')

        try:
            return backup.write('ratings.json', items)
        except Exception as ex:
            log.error('Unable to write ratings to disk: %s', ex, exc_info=True)

        return False



from __future__ import print_function

from trakt_tools.core.input import boolean_input
from trakt_tools.models import Profile
from trakt_tools.tasks.profile.backup import CreateBackupTask
from trakt_tools.tasks.base import Task
from .executor import Executor
from ..scan import ScanHistoryDuplicatesTask

from trakt import Trakt
import logging

log = logging.getLogger(__name__)


class MergeHistoryDuplicatesTask(Task):
    def __init__(self, backup_dir, delta_max, per_page=1000, debug=False, rate_limit=20):
        super(MergeHistoryDuplicatesTask, self).__init__(
            debug=debug,
            rate_limit=rate_limit
        )

        self.backup_dir = backup_dir
        self.delta_max = delta_max
        self.per_page = per_page

        self.scanner = None

    def run(self, token, backup=None, review=None):
        log.debug('run()')

        # Process backup download
        with Trakt.configuration.oauth(token=token):
            return self.process(
                backup=backup,
                review=review
            )

    def process(self, profile=None, backup=None, review=None):
        log.debug('process()')

        if not profile:
            print('Requesting profile...')
            profile = Profile.fetch(
                self.per_page,
                self.rate_limit
            )

        if not profile:
            print('Unable to fetch profile')
            exit(1)

        print('Logged in as %r' % profile.username)
        print()

        if not boolean_input('Would you like to continue?', default=True):
            exit(0)

        print()

        # Create backup
        if backup is None:
            backup = boolean_input('Create profile backup?', default=True)
            print()

        if backup and not self._create_backup(profile):
            print('Unable to create backup')
            exit(1)

        print()

        # Construct new duplicate scanner
        self.scanner = ScanHistoryDuplicatesTask(
            delta_max=self.delta_max,

            debug=self.debug,
            rate_limit=self.rate_limit
        )

        # Scan history for duplicates
        if not self.scanner.scan(profile):
            exit(1)

        print()

        if len(self.scanner.shows) or len(self.scanner.movies):
            print('Found %d show(s) and %d movie(s) with duplicates' % (
                len(self.scanner.shows),
                len(self.scanner.movies)
            ))
        else:
            print('Unable to find any duplicates')
            exit(0)

        print()

        # Execute actions
        if not self.execute(profile, review):
            print('Unable to execute actions')
            exit(1)

        return True

    def execute(self, profile, review=None):
        if review is None:
            review = boolean_input('Review every action?', default=True)
            print()

        executor = Executor(review)

        if not executor.process_shows(profile, self.scanner.shows):
            return False

        if not executor.process_movies(profile, self.scanner.movies):
            return False

        print('Done')
        return True

    # region Private methods

    def _create_backup(self, profile):
        return CreateBackupTask(
            backup_dir=self.backup_dir,

            debug=self.debug,
            rate_limit=self.rate_limit
        ).create_backup(
            profile
        )

    # endregion



from __future__ import print_function

from trakt_tools.core.input import boolean_input
from ..core.formatter import Formatter

from requests import RequestException
from trakt import Trakt, ClientError, ServerError
import logging
import six

log = logging.getLogger(__name__)


class Executor(object):
    def __init__(self, review=True, batch_size=200):
        self.review = review
        self.batch_size = batch_size

    def process_shows(self, profile, shows):
        log.debug('Executing actions on %d shows...', len(shows))

        timezone = profile.timezone

        for _, show in shows.items():
            if not show.children:
                continue

            title, ids = Formatter.show(show, timezone=timezone)
            print()

            # Review actions
            if self.review and not boolean_input(
                'Remove %d duplicate history record(s) for %s?' % (len(ids), title),
                default=True
            ):
                print('Skipped')
                continue

            # Remove history records
            for x in six.moves.xrange(0, len(ids), self.batch_size):
                self._remove_records(ids[x:x + self.batch_size])

            print()
            print('-' * 70)
            print()

        return True

    def process_movies(self, profile, movies):
        log.debug('Executing actions on %d movies...', len(movies))

        timezone = profile.timezone

        for _, movie in movies.items():
            title, ids = Formatter.movie(movie, timezone=timezone)
            print()

            # Review actions
            if self.review and not boolean_input(
                'Remove %d duplicate history record(s) for %s?' % (len(ids), title),
                default=True
            ):
                print('Skipped')
                continue

            # Remove history records
            for x in six.moves.xrange(0, len(ids), self.batch_size):
                self._remove_records(ids[x:x + self.batch_size])

            print()
            print('-' * 70)
            print()

        return True

    @staticmethod
    def _remove_records(ids):
        while True:
            # Attempt removal of records
            try:
                response = Trakt['sync/history'].remove(
                    items={
                        'ids': ids
                    },
                    exceptions=True
                )
            except (ClientError, ServerError, RequestException) as ex:
                # Retrieve error message
                message = str(ex)

                if isinstance(ex, (ClientError, ServerError)):
                    _, message = ex.error
                elif hasattr(ex, 'message'):
                    message = ex.message

                # Prompt for request retry
                print('Unable to remove %d history record(s): %s' % (len(ids), message))

                if not boolean_input('Would you like to retry the request?', default=True):
                    # Cancel request
                    return False

                # Retry request
                print()
                continue

            # Display results
            deleted_episodes = response.get('deleted', {}).get('episodes', 0)
            deleted_movies = response.get('deleted', {}).get('movies', 0)

            if deleted_episodes and deleted_movies:
                print('Removed %d episode record(s) and %d movie record(s) from history' % (
                    deleted_episodes,
                    deleted_movies
                ))
            elif deleted_episodes:
                print('Removed %d episode record(s) from history' % (
                    deleted_episodes
                ))
            elif deleted_movies:
                print('Removed %d movie record(s) from history' % (
                    deleted_movies
                ))

            for record_id in response.get('not_found', {}).get('ids', []):
                print('Unable to find record with id: %r' % record_id)

            return True



# from .collection import CollectionHandler
from .history import HistoryHandler
# from .playback import PlaybackHandler
# from .ratings import RatingsHandler
# from .watchlist import WatchlistHandler


__all__ = [
    # 'CollectionHandler',
    'HistoryHandler',
    # 'PlaybackHandler',
    # 'RatingsHandler',
    # 'WatchlistHandler'
]



from .history import history_duplicates_merge, history_duplicates_scan
from .profile import profile_backup_apply, profile_backup_create


__all__ = [
    'history_duplicates_merge',
    'history_duplicates_scan',
    'profile_backup_apply',
    'profile_backup_create'
]



from __future__ import print_function

from trakt import Trakt
import six


def authenticate():
    # Request authentication
    print('Navigate to %s' % Trakt['oauth/pin'].url())
    pin = six.moves.input('Pin: ')

    # Exchange `code` for `access_token`
    authorization = Trakt['oauth'].token_exchange(pin, 'urn:ietf:wg:oauth:2.0:oob')

    if not authorization or not authorization.get('access_token'):
        return False, None

    # Print token so user doesn't have to create PIN each time
    print("Token:", authorization['access_token'])

    return True, authorization['access_token']



# flake8: noqa: W504
from __future__ import print_function

import logging
import os

log = logging.getLogger(__name__)


class CollectionHandler(object):
    def run(self, backup, profile):
        print('Collection')

        return (
            self.run_media(backup, profile, 'movies') and
            self.run_media(backup, profile, 'shows')
        )

    def run_media(self, backup, profile, media):
        # Request ratings
        response = profile.get('/sync/collection/%s?extended=metadata' % media)

        if response.status_code != 200:
            print('Invalid response returned')
            return False

        # Retrieve items
        items = response.json()

        if media == 'movies':
            print(' - Received %d movie(s)' % len(items))
        elif media == 'shows':
            print(' - Received %d show(s)' % len(items))
        else:
            print(' - Received %d item(s)' % len(items))

        # Ensure collection directory exists
        collection_dir = os.path.join(backup.path, 'collection')

        if not os.path.exists(collection_dir):
            os.makedirs(collection_dir)

        # Write collected items to disk
        dest_path = os.path.join('collection', '%s.json' % media)

        print(' - Writing to "%s"...' % dest_path)

        try:
            return backup.write(dest_path, items)
        except Exception as ex:
            log.error('Unable to write collected items to disk: %s', ex, exc_info=True)

        return False



from __future__ import print_function

import logging

log = logging.getLogger(__name__)


class HistoryHandler(object):
    def run(self, backup, profile):
        print('History')

        items = []

        for i, count, page in profile.get_pages('/sync/history'):
            # Append `page` items to list
            items.extend(page)

            print(' - Received %d item(s) (page %d of %d)' % (
                len(page),
                i, count
            ))

        # Write watched history to disk
        print(' - Writing to "history.json"...')

        try:
            return backup.write('history.json', items)
        except Exception as ex:
            log.error('Unable to write watched history to disk: %s', ex, exc_info=True)

        return False



from trakt_tools.core.authentication import authenticate
from trakt_tools.tasks import CreateApplyTask

import click
import os


@click.command('profile:backup:apply', short_help="Apply backup to a Trakt.tv profile (history)")
@click.argument('backup_zip', type=click.Path(exists=True))
@click.option(
    '--token',
    default=os.environ.get('TRAKT_TOKEN') or None,
    help='Trakt.tv authentication token. (default: "TRAKT_TOKEN" or Prompt)'
)
def profile_backup_apply(backup_zip, token):
    """Apply backup to a Trakt.tv profile.

    Only history can be applied to your profile currently. Support for applying collection, playback, ratings, and
    watchlist data has not been implemented yet.

    Note: History already on your profile will be duplicated, `history:duplicates:merge` can be run afterwards to merge
    any duplicates in your history.

    BACKUP_ZIP is the location of the zip file created by the profile:history:backup command
    """

    if not token:
        success, token = authenticate()

        if not success:
            print('Authentication failed')
            exit(1)

        print()

    # Ensure backup directory exists
    if not os.path.exists(backup_zip):
        print('No such backup zip: "{}"'.format(backup_zip))
        exit(1)

    # Run task
    success = CreateApplyTask(backup_zip).run(token=token)

    if not success:
        exit(1)



from __future__ import print_function

from trakt_tools.core.authentication import authenticate
from trakt_tools.tasks import ScanHistoryDuplicatesTask

import click
import os


@click.command('history:duplicates:scan')
@click.option(
    '--token',
    default=os.environ.get('TRAKT_TOKEN') or None,
    help='Trakt.tv authentication token. (default: "TRAKT_TOKEN" or Prompt)'
)
@click.option(
    '--delta-max',
    default=10 * 60,
    help='Maximum delta between history records to consider as duplicate. (in seconds) (default: 600)'
)
@click.option(
    '--per-page',
    default=1000,
    help='Request page size. (default: 1000)'
)
@click.pass_context
def history_duplicates_scan(ctx, token, delta_max, per_page):
    """Scan for duplicate history records"""

    if not token:
        success, token = authenticate()

        if not success:
            print('Authentication failed')
            exit(1)

        print()

    # Run task
    success = ScanHistoryDuplicatesTask(
        delta_max=delta_max,
        per_page=per_page,

        debug=ctx.parent.debug,
        rate_limit=ctx.parent.rate_limit
    ).run(
        token=token
    )

    if not success:
        exit(1)



import six


def boolean_input(message, default=None):
    while True:
        result = six.moves.input('%s [%s]: ' % (message, 'yes' if default else 'no'))

        # Strip whitespace and convert to lower case
        result = result.strip().lower()

        # Parse result
        if result == '':
            return default

        if result in ['yes', 'y']:
            return True

        if result in ['no', 'n']:
            return False



from trakt_tools.tasks import ScanHistoryDuplicatesTask

from datetime import datetime
from dateutil.tz import tzutc


#
# Single duplicates
#


def test_duplicate_single_episode():
    scanner = ScanHistoryDuplicatesTask(
        delta_max=600
    )

    # Process history
    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'episode',

        'id': 1,
        'watched_at': '2014-03-31T09:28:53.000Z',

        'episode': {
            'season': 2,
            'number': 1,
            'title': 'Pawnee Zoo',

            'ids': {
                'trakt': 251,
                'tvdb': 797571,
                'tmdb': 397629
            }
        },

        'show': {
            'title': 'Parks and Recreation',
            'year': 2009,

            'ids': {
                'trakt': 4,
                'slug': 'parks-and-recreation',
                'tvdb': 84912,
                'imdb': 'tt1266020',
                'tmdb': 8592,
                'tvrage': 21686
            }
        }
    }) is True

    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'episode',

        'id': 2,
        'watched_at': '2014-03-31T09:20:53.000Z',

        'episode': {
            'season': 2,
            'number': 1,
            'title': 'Pawnee Zoo',

            'ids': {
                'trakt': 251,
                'tvdb': 797571,
                'tmdb': 397629
            }
        },

        'show': {
            'title': 'Parks and Recreation',
            'year': 2009,

            'ids': {
                'trakt': 4,
                'slug': 'parks-and-recreation',
                'tvdb': 84912,
                'imdb': 'tt1266020',
                'tmdb': 8592,
                'tvrage': 21686
            }
        }
    }) is True

    # Close scanner
    scanner.close()

    # Validate entry
    show = scanner.shows.get('4')
    assert show is not None

    assert show.title == 'Parks and Recreation'
    assert show.year == 2009

    # Validate children
    assert len(show.children) == 1

    episode = show.children.get('251')
    assert episode is not None

    # Validate groups
    assert len(episode.groups) == 1

    assert [
        r.id for r in episode.groups.get(datetime(2014, 3, 31, 9, 28, 53, tzinfo=tzutc()), [])
    ] == [
        1,
        2
    ]

    # Validate records
    assert sorted(episode.records.keys()) == [
        1,
        2
    ]


def test_duplicate_single_movie():
    scanner = ScanHistoryDuplicatesTask(
        delta_max=600
    )

    # Process history
    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'movie',

        'id': 1,
        'watched_at': '2014-03-31T09:28:53.000Z',

        'movie': {
            'title': 'The Dark Knight',
            'year': 2008,

            'ids': {
                'trakt': 4,
                'slug': 'the-dark-knight-2008',
                'imdb': 'tt0468569',
                'tmdb': 155
            }
        }
    }) is True

    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'movie',

        'id': 2,
        'watched_at': '2014-03-31T09:20:53.000Z',

        'movie': {
            'title': 'The Dark Knight',
            'year': 2008,

            'ids': {
                'trakt': 4,
                'slug': 'the-dark-knight-2008',
                'imdb': 'tt0468569',
                'tmdb': 155
            }
        }
    }) is True

    # Close scanner
    scanner.close()

    # Validate entry
    movie = scanner.movies.get('4')
    assert movie is not None

    assert movie.title == 'The Dark Knight'
    assert movie.year == 2008

    # Validate groups
    assert len(movie.groups) == 1

    assert [
        r.id for r in movie.groups.get(datetime(2014, 3, 31, 9, 28, 53, tzinfo=tzutc()), [])
    ] == [
        1,
        2
    ]

    # Validate records
    assert sorted(movie.records.keys()) == [
        1,
        2
    ]


#
# Unique
#


def test_unique_episode():
    scanner = ScanHistoryDuplicatesTask(
        delta_max=600
    )

    # Process history
    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'episode',

        'id': 1,
        'watched_at': '2014-03-31T09:28:53.000Z',

        'episode': {
            'season': 2,
            'number': 1,
            'title': 'Pawnee Zoo',

            'ids': {
                'trakt': 251,
                'tvdb': 797571,
                'tmdb': 397629
            }
        },

        'show': {
            'title': 'Parks and Recreation',
            'year': 2009,

            'ids': {
                'trakt': 4,
                'slug': 'parks-and-recreation',
                'tvdb': 84912,
                'imdb': 'tt1266020',
                'tmdb': 8592,
                'tvrage': 21686
            }
        }
    }) is True

    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'episode',

        'id': 2,
        'watched_at': '2014-03-30T09:28:53.000Z',

        'episode': {
            'season': 2,
            'number': 1,
            'title': 'Pawnee Zoo',

            'ids': {
                'trakt': 251,
                'tvdb': 797571,
                'tmdb': 397629
            }
        },

        'show': {
            'title': 'Parks and Recreation',
            'year': 2009,

            'ids': {
                'trakt': 4,
                'slug': 'parks-and-recreation',
                'tvdb': 84912,
                'imdb': 'tt1266020',
                'tmdb': 8592,
                'tvrage': 21686
            }
        }
    }) is True

    # Close scanner
    scanner.close()

    # Validate entry
    assert scanner.shows.get('4') is None


def test_unique_movie():
    scanner = ScanHistoryDuplicatesTask(
        delta_max=600
    )

    # Process history
    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'movie',

        'id': 1,
        'watched_at': '2014-03-30T09:28:53.000Z',

        'movie': {
            'title': 'The Dark Knight',
            'year': 2008,

            'ids': {
                'trakt': 4,
                'slug': 'the-dark-knight-2008',
                'imdb': 'tt0468569',
                'tmdb': 155
            }
        }
    }) is True

    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'movie',

        'id': 2,
        'watched_at': '2014-03-31T09:28:53.000Z',

        'movie': {
            'title': 'The Dark Knight',
            'year': 2008,

            'ids': {
                'trakt': 4,
                'slug': 'the-dark-knight-2008',
                'imdb': 'tt0468569',
                'tmdb': 155
            }
        }
    }) is True

    # Close scanner
    scanner.close()

    # Validate entry
    assert scanner.movies.get('4') is None


#
# Record duplication
#


def test_record_duplication():
    scanner = ScanHistoryDuplicatesTask(
        delta_max=600
    )

    # Process history
    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'movie',

        'id': 1,
        'watched_at': '2014-03-31T09:28:53.000Z',

        'movie': {
            'title': 'The Dark Knight',
            'year': 2008,

            'ids': {
                'trakt': 4,
                'slug': 'the-dark-knight-2008',
                'imdb': 'tt0468569',
                'tmdb': 155
            }
        }
    }) is True

    assert scanner.process_item({
        'action': 'scrobble',
        'type': 'movie',

        'id': 1,
        'watched_at': '2014-03-31T09:20:53.000Z',

        'movie': {
            'title': 'The Dark Knight',
            'year': 2008,

            'ids': {
                'trakt': 4,
                'slug': 'the-dark-knight-2008',
                'imdb': 'tt0468569',
                'tmdb': 155
            }
        }
    }) is True

    # Close scanner
    scanner.close()

    # Validate entry
    assert scanner.movies.get('4') is None



from __future__ import print_function

import six


class Formatter(object):
    @classmethod
    def movie(cls, movie, timezone=None):
        title = '"%s" (%r)' % (
            movie.title,
            movie.year
        )

        print('%s - %d plays -> %d plays' % (
            title,
            len(movie.records),
            len(movie.groups)
        ))

        ids = []

        for timestamp_utc, records in movie.groups.items():
            if timezone:
                timestamp = timestamp_utc.astimezone(timezone)
            else:
                timestamp = timestamp_utc

            print('\t%s (%s)' % (
                timestamp.strftime('%b %d, %Y %I:%M %p %Z'),
                timestamp_utc.isoformat()
            ))

            ids.extend([
                record.id for record in records[1:]
            ])

        return title, ids

    @classmethod
    def show(cls, show, timezone=None):
        title = '"%s" (%r)' % (
            show.title,
            show.year
        )

        print('%s' % title)

        return title, Formatter.episodes(show, timezone=timezone)

    @classmethod
    def episodes(cls, show, timezone=None):
        ids = []

        for x, episode in enumerate(six.itervalues(show.children)):
            print('\tS%02dE%02d - %d plays -> %d plays' % (
                episode.season,
                episode.number,
                len(episode.records),
                len(episode.groups)
            ))

            for timestamp_utc, records in episode.groups.items():
                if timezone:
                    timestamp = timestamp_utc.astimezone(timezone)
                else:
                    timestamp = timestamp_utc

                print('\t\t%s (%s)' % (
                    timestamp.strftime('%b %d, %Y %I:%M %p %Z'),
                    timestamp_utc.isoformat()
                ))

                ids.extend([
                    record.id for record in records[1:]
                ])

            if x < len(show.children) - 1:
                print()

        return ids



from datetime import datetime
import json
import os


class Backup(object):
    def __init__(self, name, path, username=None, timestamp=None):
        self.name = name
        self.path = path

        self.username = username
        self.timestamp = timestamp

    #
    # Public methods
    #

    def write(self, path, data):
        path = os.path.join(self.path, path)

        if os.path.exists(path):
            return False

        with open(path, 'w') as fp:
            json.dump(data, fp, indent=4, sort_keys=True)

        return True

    #
    # Static methods
    #

    @classmethod
    def create(cls, backup_dir, username, timestamp=None):
        if timestamp is None:
            timestamp = datetime.utcnow()

        # Build name
        name = timestamp.strftime("%Y-%m-%d_%H-%M")

        # Build path
        path = os.path.join(
            backup_dir,
            username,
            name
        )

        # Ensure directory exists
        os.makedirs(path)

        # Construct backup
        return cls(
            name=name,
            path=path,

            username=username,
            timestamp=timestamp
        )



from __future__ import print_function

from trakt_tools.core.input import boolean_input
from trakt_tools.models import Profile
from trakt_tools.tasks.base import Task
from ..core.formatter import Formatter
from .models import Entry, Record

from trakt import Trakt
from trakt.mapper import SyncMapper
from trakt.objects import Episode
import logging
import six

log = logging.getLogger(__name__)


class ScanHistoryDuplicatesTask(Task):
    def __init__(self, delta_max, per_page=1000, debug=False, rate_limit=20):
        super(ScanHistoryDuplicatesTask, self).__init__(
            debug=debug,
            rate_limit=rate_limit
        )

        self.delta_max = delta_max
        self.per_page = per_page

        self.shows = {}
        self.movies = {}

        self._current_shows = {}
        self._current_movies = {}

    def run(self, token):
        log.debug('run()')

        # Process backup download
        with Trakt.configuration.oauth(token=token):
            return self.process()

    def process(self, profile=None):
        log.debug('process()')

        if not profile:
            print('Requesting profile...')
            profile = Profile.fetch(
                self.per_page,
                self.rate_limit
            )

        if not profile:
            print('Unable to fetch profile')
            exit(1)

        print('Logged in as %r' % profile.username)
        print()

        if not boolean_input('Would you like to continue?', default=True):
            exit(0)

        print()

        if not self.scan(profile):
            exit(1)

        print()

        if not self.shows and not self.movies:
            print('Unable to find any duplicates')
            exit(0)

        timezone = profile.timezone

        # Display duplicate shows
        for _, show in self.shows.items():
            if not show.children:
                continue

            Formatter.show(show, timezone=timezone)
            print()

        # Display duplicate movies
        for _, movie in self.movies.items():
            Formatter.movie(movie, timezone=timezone)
            print()

        return True

    def scan(self, profile):
        print('Scanning for duplicates...')

        # Process items
        for i, count, page in profile.get_pages('/sync/history'):
            print(' - Processing %d items... (page %d of %d)' % (len(page), i, count))

            for item in page:
                # Process item, stop scanning if an error is encountered
                if not self.process_item(item):
                    print('Unable to process item: %r' % item)
                    return False

        # Close scanner (find duplicates, release resources)
        self.close()
        return True

    def close(self):
        # Find duplicated items
        self.shows = self._get_duplicated_items(self._current_shows)
        self.movies = self._get_duplicated_items(self._current_movies)

        # Destroy item stores
        self._current_shows = None
        self._current_movies = None

    def process_item(self, data):
        if 'episode' in data:
            return self.map_item(self._current_shows, SyncMapper.episode(None, None, data))

        if 'movie' in data:
            return self.map_item(self._current_movies, SyncMapper.movie(None, None, data))

        log.warn('Unknown item: %r', data)
        return False

    def map_item(self, store, current, create_shows=True):
        # Create show structure
        if create_shows and isinstance(current, Episode):
            show_key = current.show.get_key('trakt')

            if not show_key:
                log.warn('Unable to find "trakt" key in show: %r', current.show)
                return False

            if show_key not in store:
                store[show_key] = Entry.from_item(current.show)

            return self.map_item(store[show_key].children, current, create_shows=False)

        # Retrieve item key
        key = current.get_key('trakt')

        if not key:
            log.warn('Unable to find "trakt" key in item: %r', current)
            return False

        # Check if item already exists
        if key not in store:
            # Create new entry
            store[key] = Entry.from_item(current)
            return True

        # Try add record to existing group
        record = Record.from_item(current)

        if store[key].has_record(record):
            log.debug('Ignoring duplicate record: %r', record)
            return True

        if store[key].add(record, self.delta_max):
            return True

        # Create new group
        store[key].create_group(record)
        return True

    @staticmethod
    def _get_duplicated_items(store):
        result = {}

        for key, entry in six.iteritems(store):
            if entry.duplicated:
                result[key] = entry

            if entry.children:
                # Update `entry` with duplicated children
                entry.children = dict([
                    (k, e) for k, e in entry.children.items()
                    if e.duplicated
                ])

                # Include if there is at least one duplicated child
                if entry.children:
                    result[key] = entry

        return result



import json
import logging
from zipfile import ZipFile

import six
from trakt import Trakt

from trakt_tools.core.input import boolean_input

log = logging.getLogger(__name__)


class HistoryHandler(object):
    def run(self, backup_zip, batch_size=200):
        print('Applying history')

        episodes = []
        movies = []

        with ZipFile(backup_zip, 'r') as zipFile:
            with zipFile.open('history.json', 'r') as fp:
                history = json.load(fp)

        for item in history:
            item_type = item.get('type')
            watched_at = item.get('watched_at')

            if item_type == 'movie':
                ids = item.get('movie').get('ids')
                movies.append({'watched_at': watched_at, 'ids': ids})

            elif item_type == 'episode':
                ids = item.get('episode').get('ids')
                episodes.append({'watched_at': watched_at, 'ids': ids})

            else:
                print('Unable to apply {}: {}'.format(item_type, item))

        # Add the episodes in batches
        print('Adding {} Episodes in batches of {}:'.format(len(episodes), batch_size))
        for x in six.moves.xrange(0, len(episodes), batch_size):
            if not self._add_episodes(episodes[x:x + batch_size]):
                return False

        # Add the movies in batches
        print('Adding {} Movies in batches of {}:'.format(len(movies), batch_size))
        for x in six.moves.xrange(0, len(movies), batch_size):
            if not self._add_movies(movies[x:x + batch_size]):
                return False

        return True

    def _add_episodes(self, episodes):
        response = Trakt['sync/history'].add({'episodes': episodes})

        print(' - Added {} Episodes'.format(response.get("added").get("episodes")))

        failed_episodes = response.get('not_found').get('episodes')
        if len(failed_episodes) > 0:
            print()
            print('Unable to apply Episodes:')
            for episode in failed_episodes:
                print(' - ', episode)

            print()
            if not boolean_input('Would you like to continue anyway?', default=False):
                return False

        return True

    def _add_movies(self, movies):
        response = Trakt['sync/history'].add({'movies': movies})

        print(' - Added {} Movies'.format(response.get("added").get("movies")))

        failed_movies = response.get('not_found').get('movies')
        if len(failed_movies) > 0:
            print()
            print('Unable to apply Movies:')
            for movie in failed_movies:
                print(' - ', movie)

            print()
            if not boolean_input('Would you like to continue anyway?', default=False):
                return False

        return True



from .history import MergeHistoryDuplicatesTask, ScanHistoryDuplicatesTask
from .profile import CreateBackupTask, CreateApplyTask


__all__ = [
    'MergeHistoryDuplicatesTask',
    'ScanHistoryDuplicatesTask',
    'CreateBackupTask',
    'CreateApplyTask'
]



from trakt_tools.core.input import boolean_input
from trakt_tools.models import Profile
from trakt_tools.tasks.base import Task
from .handlers import HistoryHandler

from trakt import Trakt

import logging

log = logging.getLogger(__name__)


class CreateApplyTask(Task):
    # TODO: implement other handlers
    handlers = [
        # CollectionHandler,
        HistoryHandler,
        # PlaybackHandler,
        # RatingsHandler,
        # WatchlistHandler
    ]

    def __init__(self, backup_zip, per_page=1000, debug=False, rate_limit=20):
        super(CreateApplyTask, self).__init__(debug=debug, rate_limit=rate_limit)

        self.backup_zip = backup_zip
        self.per_page = per_page

    def run(self, token):
        log.debug('run()')

        # Process applying backup
        with Trakt.configuration.oauth(token=token):
            return self.process()

    def process(self, profile=None):
        log.debug('process()')

        if not profile:
            print('Requesting profile...')
            profile = Profile.fetch(
                self.per_page,
                self.rate_limit
            )

        if not profile:
            raise Exception('Unable to fetch profile')

        print('Logged in as %r' % profile.username)
        print()

        if not boolean_input('Would you like to continue?', default=True):
            exit(0)

        print()

        # Apply backup
        return self.apply_backup()

    def apply_backup(self):
        for handler in self.handlers:
            h = handler()

            if not h.run(self.backup_zip):
                print('Unable to apply backup, handler %r failed' % h)
                return False

            print()



from trakt_tools.core.helpers import total_seconds

from trakt.objects import Show, Episode, Movie
import six


class Entry(object):
    __slots__ = (
        'key',
        'title',
        'year',

        'season',
        'number',

        'children',

        'groups',
        'records'
    )

    def __init__(self, key, title, year, season=None, number=None):
        self.key = key
        self.title = title
        self.year = year

        self.season = season
        self.number = number

        self.children = {}

        self.groups = {}
        self.records = {}

    @property
    def duplicated(self):
        return len(self.records) > len(self.groups)

    def add(self, current, delta_max):
        for records in six.itervalues(self.groups):
            matched = False

            for record in records:
                delta = abs(total_seconds(record.watched_at - current.watched_at))

                if delta < delta_max:
                    matched = True
                    break

            if matched:
                # Add record to group
                records.append(current)

                # Add record to list
                self.records[current.id] = current
                return True

        return False

    def create_group(self, record):
        if record.watched_at in self.groups:
            raise Exception('Group %r already exists' % record.watched_at)

        # Create record group
        self.groups[record.watched_at] = [record]

        # Add record to list
        self.records[record.id] = record

    def has_record(self, record):
        return record.id in self.records

    def __repr__(self):
        fragments = []

        # Title + Year
        fragments.append('%r (%r)' % (
            self.title,
            self.year
        ))

        # Season + Episode
        if self.season is not None and self.number is not None:
            fragments.append('S%02dE%02d' % (
                self.season,
                self.number
            ))

        # Children
        if self.children:
            fragments.append('%d children' % (
                len(self.children)
            ))

        # Groups + Records
        if self.groups or self.records:
            fragments.append('%d group(s) / %d record(s)' % (
                len(self.groups),
                len(self.records)
            ))

        return '<Entry %s>' % ' - '.join(fragments)

    @classmethod
    def from_item(cls, item):
        # Construct entry
        if isinstance(item, Episode):
            season, number = item.pk

            entry = cls(
                item.get_key('trakt'),
                item.show.title,
                item.show.year,
                season,
                number
            )
        elif isinstance(item, (Show, Movie)):
            entry = cls(
                item.get_key('trakt'),
                item.title,
                item.year
            )
        else:
            raise ValueError('Unsupported item: %r' % item)

        # Create record group
        if isinstance(item, (Episode, Movie)):
            entry.create_group(Record.from_item(item))

        return entry


class Record(object):
    __slots__ = (
        'id',
        'watched_at'
    )

    def __init__(self, id, watched_at):
        self.id = id
        self.watched_at = watched_at

    def __repr__(self):
        return '<Record id: %r, watched_at: %r>' % (
            self.id,
            self.watched_at
        )

    @classmethod
    def from_item(cls, item):
        if isinstance(item, (Episode, Movie)):
            return cls(
                item.id,
                item.watched_at
            )

        raise ValueError('Unsupported item: %r' % item)



from . import commands

from trakt import Trakt
import click
import logging
import os
import sys


# Configure trakt.py
# TODO use the trakt-tools api key
Trakt.configuration.defaults.app(
    id='10248'
)

Trakt.configuration.defaults.client(
    id='023e5e78690b1d8014b40ecff062f02f9fbe187649920b06a5e56939b4514ddf',
    secret='30bb522275beb2710334cf4db18ca157b5fc85d1e17c913229bf853a9390a1ea'
)


# Initialize command-line parser
@click.group()
@click.option('--debug/--no-debug', help='Display debug messages.')
@click.option('--rate-limit', default=20, help='Maximum number of requests per minute. (default: 20)')
@click.pass_context
def cli(ctx, debug, rate_limit):
    # Setup logging level
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARN
    )

    # Update context
    ctx.debug = debug
    ctx.rate_limit = rate_limit


# Add commands
cli.add_command(commands.profile_backup_apply)
cli.add_command(commands.profile_backup_create)
cli.add_command(commands.history_duplicates_merge)
cli.add_command(commands.history_duplicates_scan)


def get_prog():
    try:
        if os.path.basename(sys.argv[0]) in ('__main__.py', '-c'):
            return '%s -m trakt_tools' % sys.executable

        if os.path.basename(sys.argv[0]) == 'run.py':
            return 'run.py'
    except (AttributeError, TypeError, IndexError):
        pass

    return 'trakt_tools'


def main():
    cli(prog_name=get_prog(), obj={}, max_content_width=100)


if __name__ == '__main__':
    main()



from __future__ import print_function

from trakt_tools.core.authentication import authenticate
from trakt_tools.tasks import CreateBackupTask

import click
import os


@click.command('profile:backup:create')
@click.option(
    '-y', '--yes', 'assume_yes',
    is_flag=True,
    help='Automatic yes to confirmation prompts.'
)
@click.option(
    '--token',
    default=os.environ.get('TRAKT_TOKEN') or None,
    help='Trakt.tv authentication token. (default: "TRAKT_TOKEN" or Prompt)'
)
@click.option(
    '--backup-dir',
    default=None,
    help='Directory that backups should be stored in. (default: "./backups")'
)
@click.option(
    '--per-page',
    default=1000,
    help='Request page size. (default: 1000)'
)
@click.pass_context
def profile_backup_create(ctx, assume_yes, token, backup_dir, per_page):
    """Create backup of a Trakt.tv profile."""

    if not token:
        success, token = authenticate()

        if not success:
            print('Authentication failed')
            exit(1)

        print()

    # Set default backup directory
    if not backup_dir:
        backup_dir = os.path.join(os.curdir, 'backups')

    # Ensure backup directory exists
    if not os.path.exists(backup_dir):
        os.makedirs(backup_dir)

    # Run task
    success = CreateBackupTask(
        backup_dir=backup_dir,
        per_page=per_page,

        assume_yes=assume_yes,
        debug=ctx.parent.debug,
        rate_limit=ctx.parent.rate_limit
    ).run(
        token=token
    )

    if not success:
        exit(1)
